/*-----------------------------------------------*/
/*       Lingyun Jin  09-26,1998.                */
/*       Jackson state University                */
/*       Computer science Department             */
/*       Assignment 2 for CPS-615.               */
/*                                               */
/*-----------------------------------------------*/

#include  <stdio.h>
#include  <math.h>
#include  <mpi.h>


double exact_val_of_Pi;
double intg_f(double);                        /* Integral function */
double Trape(double,double,int,double);       /* Trapezoid rule    */

void main(int argc, char *argv[])
{
  int P;                       /* Number of processors */
  int my_rank;                 /* Processor rank       */
  double  a=0.0;               /* left endpoint        */
  double b=1.0;                /* right endpoint       */
  int n=1200;                  /* Number of trapezoids */
  double h;                    /* Trapezoid base length */
  
  double local_a; 
  double local_b;
  double local_n;
  
  double my_integral;          /* integral over my interval       */
  double total=0.0 ;           /* total integral, initial to 0    */
  double tmp[10];               /* temp array to store gather value */ 
  int i;
 
   /* Start up MPI */
  MPI_Init(&argc, &argv);

  /* Find out how many processes are being used. */
  MPI_Comm_size(MPI_COMM_WORLD, &P);
 
                               
  /* Get my process rank */
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
                              
  h=(b-a)/n;                 /* base length is same to all prosess  */
  local_n=n/P;               /* local number of trpezoids           */
                               
  /*  Each process's interval has length = local_n *h.    */
  /*  So, my interval starts at :                         */

  local_a = a + my_rank*local_n*h;
  local_b = local_a + local_n*h;

  my_integral= Trape(local_a, local_b, local_n, h);
  printf("Before broadcast :\n");
  printf("my rank:%d,\tlocal integral:%f\n",my_rank,my_integral);     
          
/* Gather the integrals calculated by each process        */
 
 MPI_Gather(&my_integral,1,MPI_DOUBLE,tmp,1,MPI_DOUBLE,0,MPI_COMM_WORLD);  
 if (my_rank==0)
    {
      for(i = 0; i <P; i++)
	total = total+ tmp[i];
      printf("-------------------------------------------------\n");
      printf("The computed Pi of the integral for %d grid points is  %25.16e \n",
	     n, total);

      exact_val_of_Pi = 4.0 * atan(1.0);       /* Exact value of PI   */
      printf("exact Pi =  %25.16e\n", exact_val_of_Pi);
      printf("The error or the discrepancy between exact and computed value of Pi : %25.16e\n",
	     fabs(total - exact_val_of_Pi));
      printf("-------------------------------------------------\n");
    }


  /* Broadcast the final result to other process, print the PI out   */

  MPI_Bcast(&total,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

 if (my_rank!=0)
   printf("After broadcast, my process :%d, print out pi : %25.16e\n", my_rank, total);
  
  MPI_Finalize();
}

/* end of main */



 double intg_f(double x)
 {
  return (4.0/(1.0 + x * x));     /* compute and return value */
 }


 double Trape(double local_a, double local_b, int local_n, double h)
 {
  double integral=0;               /* store loacal result in integral  */
  double x;
 
  for(x=local_a;x<local_b; x+=h)
   integral=integral+intg_f(x);
   
  integral = integral*h;
  return integral;
 } 
/* end of trapezoid rule */
